#!/usr/bin/env python
# coding: utf-8

# core
import sys, argparse
from datetime import datetime
# image recognition
from pathlib import Path
# Custom utilities
try:
    from helpers import *
    from imagetools import *
except ModuleNotFoundError:
    from MCF.helpers import *
    from MCF.imagetools import *

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def main():
    print("""
   __  ___   _          _               
  /  |/  /  (_)  ___   (_)              
 / /|_/ /  / /  / _ \ / /               
/_/__/_/  /_/  /_//_//_/                
 / ___/ ___   ___  / /_ ___  __ __  ____
/ /__  / _ \ / _ \/ __// _ \/ // / / __/
\___/__\___//_//_/\__/ \___/\_,_/ /_/   
  / __/  (_)  ___  ___/ / ___   ____    
 / _/   / /  / _ \/ _  / / -_) / __/    
/_/    /_/  /_//_/\_,_/  \__/ /_/       
                                        
""")
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", metavar='', type=str,
                    help="filepath to query image")
    parser.add_argument("-o", "--output_dir", metavar='', type=str, 
                    default='./', help="path to output directory. Default='./'")
    parser.add_argument("-p", "--prefix", metavar='', type=str,
                    action="store", dest="prefix",
                    help="new prefix for output files. By default the new files will reflect the input file's basename")
    parser.add_argument("-D", "--detectScaleBar", metavar='', type=bool,    
                    default=False, action="store", dest="detectScaleBar",
                    help="automated scale bar detection. Default=False")
    parser.add_argument("-d", "--dpi", metavar='', type=int, default=300,
                    action="store", dest="dpi",
                    help="Output image resolution in pixels. Default=300")
    parser.add_argument("--debug", type=bool, default=False,
                    action="store", dest="debug",
                    help="writes debugging information and plots more steps")
    parser.add_argument("-n", "--neighborhood", metavar='', type=int,
                    default=10, action="store", dest="neighborhood",
                    help="neighborhood size in pixels determining a unique contour. Default=10")
    parser.add_argument("-kb", "--k_blur", metavar='', type=int, 
                    default=9, action="store", dest="k_blur",
                    help="blur kernel size; must be odd. Default=9")
    parser.add_argument("-c", "--C", metavar='', type=int, default=3,
                    action="store", dest="C",
                    help="constant subtracted from mean during adaptive Gaussian smoothing. Default=3")
    parser.add_argument("-B", "--blocksize", metavar='', type=int, 
                    default=15, action="store", dest="blocksize",
                    help="neighborhood size for calculating adaptive Gaussian threshold; must be odd. Default=15")
    parser.add_argument("-kl", "--k_laplacian", metavar='', type=int, 
                    default=5, action="store", dest="k_laplacian",
                    help="Laplacian kernel size; must be odd. Default=5")
    parser.add_argument("-kd", "--k_dilate", metavar='', type=int, 
                    default=5, action="store", dest="k_dilate",
                    help="dilation kernel size; must be odd. Default=5")
    parser.add_argument("-kg", "--k_gradient", metavar='', type=int, 
                    default=5, action="store", dest="k_gradient",
                    help="gradient kernel size; must be odd. Default=3")
    parser.add_argument("-kf", "--k_foreground", metavar='', type=int, 
                    default=7, action="store", dest="k_foreground",
                    help="Foregound clean up kernel size; must be odd. Default=7")
    parser.add_argument("-a", "--Amin", metavar='', type=int, 
                        default=50, action="store", dest="Amin",
                        help="Minimum contour area in pixel")
    parser.add_argument("-A", "--Amax", metavar='', type=int, 
                        default=10e6, action="store", dest="Amax",
                        help="Maximum contour area in pixels")
    args = parser.parse_args()
    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit()

    kwargs = {}
    kwargs['k_blur'] = args.k_blur
    kwargs['C'] = args.C
    kwargs['blocksize'] = args.blocksize
    kwargs['k_laplacian'] = args.k_laplacian
    kwargs['k_dilate'] = args.k_dilate
    kwargs['k_gradient'] = args.k_gradient
    kwargs['k_foreground'] = args.k_foreground

    if not args.prefix:
        prefix = Path(args.input).stem
    else:
        prefix = args.prefix

    print("[{}] command:\n\t python mcf.py --prefix {} --neighborhood {} --dpi {} --debug {} --k_blur {} --C {} --blocksize {} --k_laplacian {} --k_dilate {} --k_gradient {} --k_foreground {} --Amin {} --Amax {:d}".format(
        datetime.now().strftime('%d %b %Y %H:%M:%S'), args.prefix, args.neighborhood, args.dpi, args.debug, args.k_blur, args.C, args.blocksize, args.k_laplacian, args.k_dilate, args.k_gradient, args.k_foreground, int(args.Amin), int(args.Amax)))

    if args.input.startswith("~/"):
        input_path = Path(Path.home() / args.input[2:])
    else:
        input_path = Path(args.input).absolute().resolve()
    if args.output_dir.startswith("~/"):
        output_dir_path = Path(Path.home() / args.output_dir[2:])
    else:
        output_dir_path = Path(args.output_dir).absolute().resolve()

    print("[{}] Input file: {}".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), input_path.absolute()))
    print("[{}] Output directory: {}".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), output_dir_path.absolute()))

    suffixes = ["contour_data.csv", "contour_data.json", "tif", "noindex.tif"]
    conflicts = [output_dir_path / "{}.{}".format(prefix, suffix) for suffix in suffixes]

    if any([c.exists() for c in conflicts]):
        print("\tFound existing files of form {}.*".format(output_dir_path / prefix))
        response = query_yes_no("Would you like to DELETE and OVERWRITE?", None)
        if response==True:
            print("\tRemoving and rerunning...")
        else:
            print("\tDelete output files or change output prefix to rerun analysis")
            sys.exit()

    process_image(image_path=input_path, prefix=prefix, Amin=args.Amin, Amax=args.Amax, neighborhood=args.neighborhood, detectScaleBar=args.detectScaleBar ,output_dir=output_dir_path, dpi=args.dpi, debug=args.debug, **kwargs)

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
if __name__ == "__main__":
    start = datetime.now()
    main()
    end = datetime.now()
    print("[{}] Time elapsed: {}".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), end-start))


#!/usr/bin/env python
# coding: utf-8

# core
from functools import partial
import sys, os, argparse
import numpy as np
from datetime import datetime
from multiprocessing import Pool, set_start_method
# image recognition
import cv2
from pathlib import Path
# Custom utilities
try:
    from helpers import *
    from imagetools import *
except ModuleNotFoundError:
    from MCF.helpers import *
    from MCF.imagetools import *

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
if __name__ == "__main__":
    set_start_method('spawn')
    start = datetime.now()

    print("""
           __  ___   _          _               
          /  |/  /  (_)  ___   (_)              
         / /|_/ /  / /  / _ \ / /               
        /_/__/_/  /_/  /_//_//_/                
         / ___/ ___   ___  / /_ ___  __ __  ____
        / /__  / _ \ / _ \/ __// _ \/ // / / __/
        \___/__\___//_//_/\__/ \___/\_,_/ /_/   
          / __/  (_)  ___  ___/ / ___   ____    
         / _/   / /  / _ \/ _  / / -_) / __/    
        /_/    /_/  /_//_/\_,_/  \__/ /_/                                              
        """)

    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", metavar='', type=str,
                    help="filepath to query image")
    parser.add_argument("-o", "--output_dir", metavar='', type=str, 
                    default='./', help="path to output directory. Default='./'")
    parser.add_argument("-S", "--stepsize", metavar='',
                    type=int, action="store", dest="stepsize",
                    help="Slide step size in pixels (same in x and y directions). Default=500")
    parser.add_argument("-W", "--winW", metavar='', type=int,
                    action="store", dest="winW",
                    help="Window width in pixels. Default=1000")
    parser.add_argument("-H", "--winH", metavar='',
                    type=int, action="store", dest="winH",
                    help="Window height in pixels. Default=1000")
    parser.add_argument("--cpus", default=1,
                        action="store", dest="cpus",
                        help="Number of cpus. Will default to 1. Use 'AUTO' to detect logical cores.")
    parser.add_argument("-p", "--prefix", metavar='', type=str,
                    action="store", dest="prefix",
                    help="new prefix for output files. By default the new files will reflect the input file's basename")
    parser.add_argument("-D", "--detectScaleBar", metavar='', type=bool,    
                    default=False, action="store", dest="detectScaleBar",
                    help="automated scale bar detection.")
    parser.add_argument("-d", "--dpi", metavar='', type=int, default=300,
                    action="store", dest="dpi",
                    help="Output image resolution in pixels. Default=300")
    parser.add_argument("--debug", type=bool, default=False,
                    action="store", dest="debug",
                    help="writes debugging information and plots more steps")
    parser.add_argument("-n", "--neighborhood", metavar='', type=int,
                    default=10, action="store", dest="neighborhood",
                    help="neighborhood size in pixels determining a unique contour. Default=10")
    parser.add_argument("-kb", "--k_blur", metavar='', type=int, 
                    default=9, action="store", dest="k_blur",
                    help="blur kernel size; must be odd. Default=9")
    parser.add_argument("-c", "--C", metavar='', type=int, default=3,
                    action="store", dest="C",
                    help="constant subtracted from mean during adaptive Gaussian smoothing. Default=3")
    parser.add_argument("-B", "--blocksize", metavar='', type=int, 
                    default=15, action="store", dest="blocksize",
                    help="neighborhood size for calculating adaptive Gaussian threshold; must be odd. Default=15")
    parser.add_argument("-kl", "--k_laplacian", metavar='', type=int, 
                    default=5, action="store", dest="k_laplacian",
                    help="Laplacian kernel size; must be odd. Default=5")
    parser.add_argument("-kd", "--k_dilate", metavar='', type=int, 
                    default=5, action="store", dest="k_dilate",
                    help="dilation kernel size; must be odd. Default=5")
    parser.add_argument("-kg", "--k_gradient", metavar='', type=int, 
                    default=5, action="store", dest="k_gradient",
                    help="gradient kernel size; must be odd. Default=3")
    parser.add_argument("-kf", "--k_foreground", metavar='', type=int, 
                    default=7, action="store", dest="k_foreground",
                    help="Foregound clean up kernel size; must be odd. Default=7")
    parser.add_argument("-a", "--Amin", metavar='', type=int, 
                        default=50, action="store", dest="Amin",
                        help="Minimum contour area in pixel")
    parser.add_argument("-A", "--Amax", metavar='', type=int, 
                        default=10e6, action="store", dest="Amax",
                        help="Maximum contour area in pixels")

    args = parser.parse_args()
    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit()

    kwargs = {}
    kwargs['k_blur'] = args.k_blur
    kwargs['C'] = args.C
    kwargs['blocksize'] = args.blocksize
    kwargs['k_laplacian'] = args.k_laplacian
    kwargs['k_dilate'] = args.k_dilate
    kwargs['k_gradient'] = args.k_gradient
    kwargs['k_foreground'] = args.k_foreground

    h, w, _ = cv2.imread(args.input).shape
    if not args.winW:
        winW = round((w / 5) / 100) * 100
    else:
        winW = args.winW
    if not args.winH:
        winH = round((h / 5) / 100) * 100
    else:
        winH = args.winH
    if not args.stepsize:
        stepsize = int(min([winH, winW])/2)
    else:
        stepsize = args.stepsize
    if not args.prefix:
        prefix = Path(args.input).stem
    else:
        prefix = args.prefix

    print("[{}] command:\n\t python mcf_parallel.py --prefix {} --stepsize {} --winW {} --winH {} --neighborhood {} --dpi {} --debug {} --k_blur {} --C {} --blocksize {} --k_laplacian {} --k_dilate {} --k_gradient {} --k_foreground {} --Amin {} --Amax {:d} --cpus {}".format(
        datetime.now().strftime('%d %b %Y %H:%M:%S'), args.prefix, stepsize, winW, winH, args.neighborhood, args.dpi, args.debug, args.k_blur, args.C, args.blocksize, args.k_laplacian, args.k_dilate, args.k_gradient, args.k_foreground, int(args.Amin), int(args.Amax), args.cpus))
    if args.cpus == "AUTO":
        cpus = get_cpus_avail()
    else:
        cpus = args.cpus
    print("[{}] Using {} cpus".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), cpus))

    if args.input.startswith("~/"):
        input_path = Path(Path.home() / args.input[2:])
    else:
        input_path = Path(args.input).absolute().resolve()
    if args.output_dir.startswith("~/"):
        output_dir_path = Path(Path.home() / args.output_dir[2:])
    else:
        output_dir_path = Path(args.output_dir).absolute().resolve()

    print("[{}] Input file: {}".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), input_path.absolute()))
    print("[{}] Output directory: {}".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), output_dir_path.absolute()))

    suffixes = ["contour_data.csv", "contour_data.json", "pdf", "noindex.pdf"]
    conflicts = [output_dir_path / "{}.{}".format(prefix, suffix) for suffix in suffixes]

    if any([c.exists() for c in conflicts]):
        print("\tFound existing files of form {}.*".format(output_dir_path / prefix))
        response = query_yes_no("Would you like to DELETE and OVERWRITE?", None)
        if response==True:
            print("\tRemoving and rerunning...")
        else:
            print("\tDelete output files or change output prefix to rerun analysis")
            sys.exit()

    if args.debug:
        print("[{}] Working directory: {}".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), os.getcwd()))
        print("[{}] Output directory: {}".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), output_dir_path))

    image = cv2.imread(input_path.as_posix())
    if image == None:
        image = vector2cv2(image_path.as_posix())

    if not prefix:
        prefix = input_path.stem
    if debug:
        print("[{}] Input file: {}".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), input_path.absolute()))

    """Denoise"""
    wdir = image_path.parent
    results = wdir.glob("{}.denoise.*".format(image_path.stem))
    try:
        denoised_path = next(results)
        denoise = cv2.imread(denoised_path.as_posix())
        print("[{}] Found existing denoised file ({})".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), denoised_path.as_posix()))
    except StopIteration:
        print("[{}] Denoising image...".format(datetime.now().strftime('%d %b %Y %H:%M:%S')))
        denoise = cv2.fastNlMeansDenoisingColored(image, None, 10, 10, 7, 21)
        denoised_path = Path("{}/{}.denoise{}".format(wdir.as_posix(), image_path.stem, ".png"))
        cv2.imwrite(filename=denoised_path.as_posix(), img=denoise)
        print("[{}] Created temporary denoised file ({})".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), denoised_path.as_posix()))

    print("[{}] Getting image border...".format(datetime.now().strftime('%d %b %Y %H:%M:%S')))
    border_contour = mcf(image=denoise, extract_border=True, **kwargs)

    # mask input image (leaves only the area inside the border contour)
    clone = denoise.copy()
    blank = np.zeros(clone.shape[0:2], dtype=np.uint8)
    border_mask = cv2.drawContours(blank.copy(), border_contour, 0, (255), -1)
    cutout = cv2.bitwise_and(clone, clone, mask=border_mask)

    # Break image into windows
    windows = list(sliding_window(image=cutout.copy(), stepSize=stepsize, windowSize=(winW, winH)))
    windows = [w for w in windows if w[2].sum() > 0]
    print("[{}] Finding contours...".format(datetime.now().strftime('%d %b %Y %H:%M:%S')))
    with Pool(processes=cpus) as pool:
        contours = pool.map(partial(parallel_mcf, **kwargs), windows)
    contours = flatten([c for c in contours if len(c) > 0])
    print("[{}] Size selecting contours...".format(datetime.now().strftime('%d %b %Y %H:%M:%S')))
    contours = contour_size_selection(contours, Amin=args.Amin, Amax=args.Amax)
    print("[{}] Found {} contours in all windows".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), len(contours)))
    
    print("[{}] Removing redundant contours...".format(datetime.now().strftime('%d %b %Y %H:%M:%S')))
    contours = remove_redundant_contours(contours, neighborhood=args.neighborhood)
    print("[{}] Found {} nonredundant contours".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), len(contours)))

    # Export contours
    print("[{}] Exporting contours...".format(datetime.now().strftime('%d %b %Y %H:%M:%S')))   
    if not output_dir_path.is_dir():
        output_dir_path.mkdir(parents=True, exist_ok=True)
    export_contour_data(image=image, contours=contours, conversion_factor=None,
                        units=None, prefix="{}.ALL".format(prefix), output_dir=output_dir_path)

    print("[{}] Plotting...".format(datetime.now().strftime('%d %b %Y %H:%M:%S')))
    render_contour_plots(image=image, border_contour=border_contour, contours=contours, prefix=prefix, dpi=args.dpi, output_dir=output_dir_path)

    end = datetime.now()
    print("[{}] Time elapsed: {}".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), end-start))


#!/usr/bin/env python
# coding: utf-8

import sys, os, re, subprocess, platform

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def flatten(l):
    """Flattens a list of sublists"""
    return [item for sublist in l for item in sublist]

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def query_yes_no(question, default):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_cpus_avail():
    if platform.system() == "Linux":
        return len(os.sched_getaffinity(0))
    elif platform.system() == "Darwin":
        stdout = subprocess.run(["sysctl","-n", "hw.ncpu"], capture_output=True, text=True).stdout
        return int(re.findall('[0-9]+', stdout[0])[0])
    elif platform.system() == "Windows":
        return int(os.environ['NUMBER_OF_PROCESSORS'])
    else:
        print("Could not determine the number of available cpus. Setting to 1.")
        return 1

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
class Vividict(dict):
    """Created by StackOverflow user Aaron Hall
    https://stackoverflow.com/a/19829714/6317380"""
    def __missing__(self, key):
        value = self[key] = type(self)() # retain local pointer to value
        return value                     # faster to return than dict lookup


#!/usr/bin/env python
# coding: utf-8

# Core
import sys, cv2, uuid
from datetime import datetime
import pyperclip
import numpy as np
np.warnings.filterwarnings('error', category=np.VisibleDeprecationWarning)
# PyQt5
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtWidgets import *
from PyQt5.QtGui import QPixmapCache
# Plotting
import matplotlib
if sys.platform == 'darwin':
    matplotlib.use("TkAgg")
# Custom utilities
try:
    from helpers import *
    from imagetools import *
except ModuleNotFoundError:
    from MCF.helpers import *
    from MCF.imagetools import *

class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        # Initialize window
        self.setWindowTitle('MCF - {}'.format(sys.argv[1]))
        self.resize(1000, 400)
        # Add toolbar with menu
        toolbar = QToolBar()
        self.addToolBar(toolbar)

        saveFile = QAction("&Save", self)
        saveFile.setShortcut("Ctrl+S")
        saveFile.setStatusTip("Save File")
        saveFile.triggered.connect(self.saveAll)
        toolbar.addAction(saveFile)

        openFile = QAction("&Open", self)
        openFile.setShortcut("Ctrl+O")
        openFile.setStatusTip('Open File')
        openFile.triggered.connect(self.file_open)
        toolbar.addAction(openFile)

        # Initialize widget with layout
        widget = QWidget()
        layout = QVBoxLayout(widget)
        # Set contour app as central widget
        self.contour_app = ContourApp(self)
        layout.addWidget(self.contour_app)
        self.setCentralWidget(widget)
        self.contour_app.update_plot()
        self.contour_app.viewer.updateView()

        self.show()
     
    def saveAll(self):
        saveDialog = QFileDialog.getSaveFileName(parent=self, 
        caption="File prefix",
        directory=Path(Path(sys.argv[1]).parent / Path(sys.argv[1]).stem).as_posix())

        if saveDialog[0] != '':
            if len(self.contour_app.contour_DF) == 0 :
                export_data = pd.DataFrame(np.zeros(shape=(len(self.contour_app.large_contours), 9)), 
                columns=["uuid4", "contour", "C", "kBlur", "blocksize", "kLaplacian","kDilate", "kGradient", "kForeground"], dtype=object)
                export_data["contour"] = self.contour_app.large_contours
                export_data["uuid4"] = [uuid.uuid4().hex for i in range(len(self.contour_app.large_contours))]
                export_data["C"] = self.contour_app.C.value()
                export_data["kBlur"] = self.contour_app.k_blur.value()
                export_data["blocksize"] = self.contour_app.blocksize.value()
                export_data["kLaplacian"] = self.contour_app.k_laplacian.value()
                export_data["kDilate"] = self.contour_app.k_dilate.value()
                export_data["kGradient"] = self.contour_app.k_gradient.value()
                export_data["kForeground"] = self.contour_app.k_foreground.value()
            else:
                export_data = self.contour_app.contour_DF

            export_contour_data(image=sys.argv[1], contourDF=export_data, conversion_factor=self.contour_app.unitsPerPixel, units=self.contour_app.scaleBarUnits,
                                output_dir=Path(saveDialog[0]).parent, prefix=Path(saveDialog[0]).stem)
            render_contour_plots(image=sys.argv[1], contours=export_data["contour"].values, border_contour=None, contour_thickness=self.contour_app.contour_thickness.value(),
                                output_dir=Path(saveDialog[0]).parent, prefix=Path(saveDialog[0]).stem, color=self.contour_app.highlight_color)
        else: return

    def file_open(self):
        fileFilter = 'Contour data file (*.json)'
        openDialog = QFileDialog.getOpenFileName(self, 
            caption = "Open contour data file (.json)",
            directory = Path(Path(sys.argv[1]).parent / Path(sys.argv[1]).stem).as_posix(),
            filter = fileFilter)

        if openDialog[0] != '':
            self.contour_app.select_contours.setChecked(True)
            loaded = pd.read_json(openDialog[0])
            loaded = loaded[self.contour_app.contour_DF.columns]
            loaded['contour'] = loaded['contour'].apply(np.array)
            print("[{}] Loaded {} contours from {}".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), len(loaded), openDialog[0]))
            self.contour_app.contour_DF = self.contour_app.contour_DF.append(loaded)
            self.contour_app.update_plot()
        else:return

    def closeEvent(self, event):
        """Generate 'question' dialog on clicking 'X' button in title bar.

        Reimplement the closeEvent() event handler to include a 'Question'
        dialog with options on how to proceed - Save, Close, Cancel buttons
        """
        reply = QMessageBox.question(
            self, "Message",
            "Are you sure you want to quit? Any unsaved work will be lost.",
            QMessageBox.Save | QMessageBox.Close | QMessageBox.Cancel,
            QMessageBox.Save)

        if reply == QMessageBox.Close:
            event.accept()
        elif reply == QMessageBox.Save:
            self.file_save()
        else:
            event.ignore()

class ContourApp(QWidget):
    """See the following StackOverflow post for embedding a QGraphicsView in
    another widget:
    https://stackoverflow.com/questions/35508711/how-to-enable-pan-and-zoom-in-a-qgraphicsview"""
    def __init__(self, parent=None):
        super().__init__(parent=parent)

        ### Create window with grid layout ###
        self.grid_layout = QGridLayout()
        self.setLayout(self.grid_layout)
        self.click_x, self.click_y = (0,0)
        self.setMouseTracking(True)

        ### Initialize viewer ###
        self.viewer = GraphicsView(self)
        self.grid_layout.addWidget(self.viewer, 0, 0, -1, 8) # Note that grid coords are (row, col, rowspan, colspan)
        self.viewer.updateView()
        self.viewer.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)
        self.viewer.setAlignment(QtCore.Qt.AlignTop)

        ### Add image and contours ###
        self.image_path = Path(sys.argv[1])
        image = cv2.imread(self.image_path.as_posix())
        if image is None:
            image = vector2cv2(self.image_path.as_posix())

        wdir = self.image_path.parent
        results = wdir.glob("{}.denoise.*".format(self.image_path.stem))
        try:
            self.denoised_path = next(results)
            self.cv2_image = cv2.imread(self.denoised_path.as_posix())
            print("[{}] Found existing denoised file ({})".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), self.denoised_path.as_posix()))
        except StopIteration:
            print("[{}] Denoising image...".format(datetime.now().strftime('%d %b %Y %H:%M:%S')))
            self.cv2_image = cv2.fastNlMeansDenoisingColored(image, None, 10, 10, 7, 21)
            self.denoised_path = Path("{}/{}.denoise{}".format(wdir.as_posix(), self.image_path.stem, ".png"))
            cv2.imwrite(filename=self.denoised_path.as_posix(), img=self.cv2_image)

            print("[{}] Created temporary denoised file ({})".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), self.denoised_path.as_posix()))

        # Get contours
        self.contours = mcf(image=self.cv2_image)
        self.contour_color = (255, 0, 0)
        self.highlight_color = (255, 0, 255)
        self.contour_DF = pd.DataFrame(columns=["uuid4", "contour", "C", "kBlur", "blocksize", "kLaplacian", "kDilate", "kGradient", "kForeground"], dtype=object)
        cv2.drawContours(self.cv2_image, contours=self.contours, contourIdx=-1, color=self.contour_color, thickness=3)
        if len(self.contour_DF)>0:
            cv2.drawContours(self.cv2_image, contours=self.contour_DF["contour"].values, contourIdx=-1, color=self.highlight_color, thickness=3)

        # Convert to Qimage object
        self.pixmap = cv2pixmap(self.cv2_image)
        self.viewer.setPhoto(self.pixmap)
        self.viewer.updateView()

        ### Create sliders ###
        self.k_blur = QSlider(QtCore.Qt.Horizontal)
        self.k_blur.setRange(2, 26)
        self.k_blur.setValue(5)
        self.k_blur.setTickInterval(1)
        self.k_blur.setTickPosition(QSlider.TicksBelow)
        self.blur_label = QLabel()
        self.blur_label.setText('k_blur: {}'.format(2*self.k_blur.value()-1))
        self.grid_layout.addWidget(self.blur_label, 0, 10)
        self.grid_layout.addWidget(self.k_blur, 0, 8, 1, 2)

        self.C = QSlider(QtCore.Qt.Horizontal)
        self.C.setRange(1, 25)
        self.C.setValue(3)
        self.C.setTickInterval(1)
        self.C.setTickPosition(QSlider.TicksBelow)
        self.C_label = QLabel()
        self.C_label.setText('C: {}'.format(self.C.value()))
        self.grid_layout.addWidget(self.C_label, 1, 10)
        self.grid_layout.addWidget(self.C, 1, 8, 1, 2)

        self.blocksize = QSlider(QtCore.Qt.Horizontal)
        self.blocksize.setRange(2, 200)
        self.blocksize.setValue(8)
        self.blocksize.setTickInterval(1)
        self.blocksize.setTickPosition(QSlider.TicksBelow)
        self.blocksize_label = QLabel()
        self.blocksize_label.setText('blocksize: {}'.format(2*self.blocksize.value()-1))
        self.grid_layout.addWidget(self.blocksize_label, 2, 10)
        self.grid_layout.addWidget(self.blocksize, 2, 8, 1, 2)

        self.k_laplacian = QSlider(QtCore.Qt.Horizontal)
        self.k_laplacian.setRange(2, 15)
        self.k_laplacian.setValue(3)
        self.k_laplacian.setTickInterval(1)
        self.k_laplacian.setTickPosition(QSlider.TicksBelow)
        self.laplacian_label = QLabel()
        self.laplacian_label.setText('k_laplacian: {}'.format(2*self.k_laplacian.value()-1))
        self.grid_layout.addWidget(self.laplacian_label, 3, 10)
        self.grid_layout.addWidget(self.k_laplacian, 3, 8, 1, 2)

        self.k_dilate = QSlider(QtCore.Qt.Horizontal)
        self.k_dilate.setRange(2, 52)
        self.k_dilate.setValue(3)
        self.k_dilate.setTickInterval(1)
        self.k_dilate.setTickPosition(QSlider.TicksBelow)
        self.dilate_label = QLabel()
        self.dilate_label.setText('k_dilate: {}'.format(2*self.k_dilate.value()-1))
        self.grid_layout.addWidget(self.dilate_label, 4, 10)
        self.grid_layout.addWidget(self.k_dilate, 4, 8, 1, 2)

        self.k_gradient = QSlider(QtCore.Qt.Horizontal)
        self.k_gradient.setRange(2, 52)
        self.k_gradient.setValue(2)
        self.k_gradient.setTickInterval(1)
        self.k_gradient.setTickPosition(QSlider.TicksBelow)
        self.gradient_label = QLabel()
        self.gradient_label.setText('k_gradient: {}'.format(2*self.k_gradient.value()-1))
        self.grid_layout.addWidget(self.gradient_label, 5, 10)
        self.grid_layout.addWidget(self.k_gradient, 5, 8, 1, 2)

        self.k_foreground = QSlider(QtCore.Qt.Horizontal)
        self.k_foreground.setRange(2, 15)
        self.k_foreground.setValue(4)
        self.k_foreground.setTickInterval(1)
        self.k_foreground.setTickPosition(QSlider.TicksBelow)
        self.foreground_label = QLabel()
        self.foreground_label.setText('k_foreground: {}'.format(2*self.k_foreground.value()-1))
        self.grid_layout.addWidget(self.foreground_label, 6, 10)
        self.grid_layout.addWidget(self.k_foreground, 6, 8, 1, 2)

        self.sliders = [self.k_blur, self.C, self.blocksize, self.k_laplacian, self.k_dilate, self.k_gradient, self.k_foreground]

        ### Set contour size ###
        self.Amin = QLineEdit()
        self.Amin.setPlaceholderText("Minimum contour area")
        self.setAmin = QPushButton("Enter")
        self.grid_layout.addWidget(self.Amin, 7, 8, 1, 2)
        self.grid_layout.addWidget(self.setAmin, 7, 10)

        self.Amax = QLineEdit()
        self.Amax.setPlaceholderText("Maximum contour area")
        self.setAmax = QPushButton("Enter")
        self.grid_layout.addWidget(self.Amax, 8, 8, 1, 2)
        self.grid_layout.addWidget(self.setAmax, 8, 10)

        ### Scale bar info
        self.scaleBar = None
        self.scaleBarUnits = None
        self.scaleBarPixelLength = None
        self.scaleBarUnitLength = None
        self.unitsPerPixel = None

        self.pixelLengthField = QLineEdit()
        self.pixelLengthField.setPlaceholderText("Length in pixels")
        self.grid_layout.addWidget(self.pixelLengthField, 9, 8, 1, 2)
        self.pixelLabel = QLabel("Pixels")
        self.grid_layout.addWidget(self.pixelLabel, 9, 10, 1, 1)

        self.unitLengthField = QLineEdit()
        self.unitLengthField.setPlaceholderText("Length in units")
        self.grid_layout.addWidget(self.unitLengthField, 10, 8, 1, 2)
        self.unitsLabel = QLabel("Length")
        self.grid_layout.addWidget(self.unitsLabel, 10, 10, 1, 1)

        self.setScaleButton = QPushButton("Set scale")
        self.grid_layout.addWidget(self.setScaleButton, 11, 8, 1, 1)

        self.detectScaleButton = QPushButton("Detect scale")
        self.grid_layout.addWidget(self.detectScaleButton, 11, 9, 1, 1)

        self.clearScaleButton = QPushButton("Clear scale")
        self.grid_layout.addWidget(self.clearScaleButton, 12, 8, 1, 1)

        self.drawing = False
        self.drawScaleButton = QPushButton("Draw scale bar")
        self.grid_layout.addWidget(self.drawScaleButton, 12, 9, 1, 1)

        ### Generate parameters for CLI ###
        self.CLIgenerator = QPushButton()
        self.CLIgenerator.setText("Copy CLI parameters")
        self.grid_layout.addWidget(self.CLIgenerator, 11, 10, 1, 1)

        ### Text box ###
        self.box = QPlainTextEdit()
        self.box.insertPlainText("--k_blur 9 --C 3 --blocksize 15 --k_laplacian 5 --k_dilate 5 --k_gradient 3 --k_foreground 7 --Amin 50 --Amax 10e6")

        ### Reset parameters ###
        self.resetParams = QPushButton()
        self.resetParams.setText("Reset parameters")
        self.grid_layout.addWidget(self.resetParams, 12, 10, 1, 1)

        ### Color picker ###
        self.contour_colorPicker = QPushButton()
        self.contour_colorPicker.setText("Contour color")
        self.grid_layout.addWidget(self.contour_colorPicker, 13, 8, 1, 1)

        self.highlight_colorPicker = QPushButton()
        self.highlight_colorPicker.setText("Selected color")
        self.grid_layout.addWidget(self.highlight_colorPicker, 13, 9, 1, 1)

        ### Contour thickness ###
        self.contour_thickness = QSlider(QtCore.Qt.Horizontal)
        self.contour_thickness.setRange(1, 10)
        self.contour_thickness.setValue(3)
        self.contour_thickness.setTickInterval(1)
        self.contour_thickness.setTickPosition(QSlider.TicksBelow)
        self.thickness_label = QLabel()
        self.thickness_label.setText('Contour thickness: {}'.format(self.contour_thickness.value()))
        self.grid_layout.addWidget(self.thickness_label, 14, 10)
        self.grid_layout.addWidget(self.contour_thickness, 14, 8, 1, 2)

        ### Contour selection ###
        self.select_contours = QCheckBox("Select contours")
        self.grid_layout.addWidget(self.select_contours, 13, 10, 1, 1)
        self.select_contours.stateChanged.connect(self.selectContoursChecked)

        ### Approximate polygons ###
        self.use_approxPolys = QCheckBox("Approximate polygons (epsilon)")
        self.grid_layout.addWidget(self.use_approxPolys, 15, 8, 1, 1)
        self.epsilon = QSlider(QtCore.Qt.Horizontal)
        self.epsilon.setRange(1, 20)
        self.epsilon.setValue(1)
        self.epsilon.setTickInterval(1)
        self.epsilon.setTickPosition(QSlider.TicksBelow)
        self.epsilon_label = QLabel()
        self.epsilon_label.setText('episilon: {}'.format(self.epsilon.value()))
        self.grid_layout.addWidget(self.epsilon_label, 16, 10)
        self.grid_layout.addWidget(self.epsilon, 16, 8, 1, 2)

        ### Convex hulls ###
        self.use_convexHulls = QCheckBox("Use convex hulls")
        self.grid_layout.addWidget(self.use_convexHulls, 15, 9, 1, 1)

        ### Connections ###
        # Sliders -> plots
        self.k_blur.valueChanged.connect(self.update_plot)
        self.C.valueChanged.connect(self.update_plot)
        self.blocksize.valueChanged.connect(self.update_plot)
        self.k_laplacian.valueChanged.connect(self.update_plot)
        self.k_dilate.valueChanged.connect(self.update_plot)
        self.k_gradient.valueChanged.connect(self.update_plot)
        self.k_foreground.valueChanged.connect(self.update_plot)
        self.contour_thickness.valueChanged.connect(self.update_plot)
        self.epsilon.valueChanged.connect(self.update_plot)
        # Sliders -> text
        self.k_blur.valueChanged.connect(self.update_text)
        self.C.valueChanged.connect(self.update_text)
        self.blocksize.valueChanged.connect(self.update_text)
        self.k_laplacian.valueChanged.connect(self.update_text)
        self.k_dilate.valueChanged.connect(self.update_text)
        self.k_gradient.valueChanged.connect(self.update_text)
        self.k_foreground.valueChanged.connect(self.update_text)
        self.contour_thickness.valueChanged.connect(self.update_text)
        self.epsilon.valueChanged.connect(self.update_text)
        
        # Line edits -> plots
        self.Amin.returnPressed.connect(self.update_plot)
        self.setAmin.clicked.connect(self.update_plot)
        self.Amax.returnPressed.connect(self.update_plot)
        self.setAmax.clicked.connect(self.update_plot)

        # Scale bars
        self.pixelLengthField.returnPressed.connect(self.setScale)
        self.unitLengthField.returnPressed.connect(self.setScale)
        self.detectScaleButton.clicked.connect(self.detectScale)
        self.clearScaleButton.clicked.connect(self.clearScale)
        self.drawScaleButton.clicked.connect(self.drawScaleBar)
        
        ### Misc. buttons ###
        # Generator push button
        self.CLIgenerator.clicked.connect(self.on_CLIgenerator_clicked)
        self.CLIgenerator.clicked.connect(self.copy_CLI)
        # Reset push button
        self.resetParams.clicked.connect(self.on_reset_clicked)
        self.resetParams.clicked.connect(self.update_text)
        self.resetParams.clicked.connect(self.update_plot)
        # Color picker button
        self.contour_colorPicker.clicked.connect(self.update_contour_color)
        self.highlight_colorPicker.clicked.connect(self.update_highlight_color)
        # Connect photo click and rubberband
        self.viewer.photoClicked.connect(self.photoClicked)
        self.viewer.rectChanged.connect(self.rectChange)
        self.update_plot()

    def drawScaleBar(self):
        if self.select_contours.isChecked():
            print("Cannot draw scale bar while selecting contours")
            return
        else:
            self.clearScale()
            self.drawing = True
            self.scaleBar = []

    def setScale(self):
        try:
            self.scaleBarPixelLength = eval(self.pixelLengthField)
            self.scaleBarUnitLength = eval(self.unitLengthField)
            self.unitsPerPixel = self.scaleBarUnitLength/self.scaleBarPixelLength
            print("Conversion factor: {} units/pixel".format(self.unitsPerPixel))
        except TypeError:
            print("Both number of pixels and a length must be set to create a scaling factor")
            return
        except NameError:
            print("Both the number of pixels and length must be numbers")
            return

    def detectScale(self):
        scaleBarInfo = get_scalebar_info(self.cv2_image)
        if len(scaleBarInfo) == 2:
            self.scaleBar, self.scaleBarPixelLength = scaleBarInfo
        elif len(scaleBarInfo) == 4:
            self.scaleBar, self.scaleBarPixelLength, self.scaleBarUnitLength, self.scaleBarUnits = scaleBarInfo
            self.unitsLabel.setText("{}".format(self.scaleBarUnits))

        self.pixelLengthField.setText("{}".format(self.scaleBarPixelLength))
        self.unitLengthField.setText("{}".format(self.scaleBarUnitLength))
        self.update_plot()

    def clearScale(self):
        self.scaleBar = None
        self.scaleBarPixelLength = None
        self.scaleBarUnitLength = None
        self.scaleBarUnits = None
        self.pixelLengthField.setText("Length in pixels")
        self.unitLengthField.setText("Length in units")
        self.unitsLabel.setText("Units")
        self.update_plot()

    def photoClicked(self, pos):
        if self.select_contours.isChecked():
            self.click_x, self.click_y = (pos.x(), pos.y())

            if self.viewer.button == 1: # Click addition
                for c in self.large_contours:
                    if cv2.pointPolygonTest(contour=c, pt=(self.click_x, self.click_y), measureDist=False) == 1.0:
                        self.contour_DF = self.contour_DF.append({
                            "uuid4": uuid.uuid4().hex, 
                            "contour": c, 
                            "C": self.C.value(), 
                            "kBlur": self.k_blur.value(),
                            "blocksize": self.blocksize.value(), 
                            "kLaplacian": self.k_laplacian.value(),
                            "kDilate": self.k_dilate.value(), 
                            "kGradient": self.k_gradient.value(), 
                            "kForeground": self.k_foreground.value()}, ignore_index=True)
                        break

            elif self.viewer.button == 2: # Click subtraction
                if len(self.contour_DF) > 0:
                    for i, row in self.contour_DF.iterrows():
                        if cv2.pointPolygonTest(contour=row["contour"], pt=(self.click_x, self.click_y), measureDist=False) == 1.0:
                            self.contour_DF = self.contour_DF.drop(index=i, axis=0)
                            self.contour_DF = self.contour_DF.reset_index(drop=True)
                            break            
            else: return
            self.update_plot()

        elif self.drawing:
            if len(self.scaleBar) == 0:
                self.scaleBar.append(pos.x())
                self.scaleBar.append(pos.y())
                self.scaleBarPixelLength = None
                self.update_plot()
            elif len(self.scaleBar) == 2:
                self.scaleBar.append(pos.x())
                self.scaleBar.append(pos.y())
                self.scaleBarPixelLength = np.sqrt((self.scaleBar[2]-self.scaleBar[0])**2+(self.scaleBar[3]-self.scaleBar[1])**2)
                self.pixelLengthField.setText("{}".format(self.scaleBarPixelLength))
                self.update_plot()

                _, okPressed = QInputDialog.getDouble(self, "Length in units", "Length:", 1, 0)
                if okPressed:
                    self.scaleBarUnitLength = _
                    _, okPressed = QInputDialog.getText(self, "Scale bar units", "Units:", QLineEdit.Normal, "\u03BC"+"m")
                    self.unitLengthField.setText(str(self.scaleBarUnitLength))
                    if okPressed and _!='':
                        self.scaleBarUnits = _
                        self.unitsLabel.setText(self.scaleBarUnits)
                else:        
                    self.unitLengthField.setText("Length in units")
                    self.unitsLabel.setText("Units")
                self.drawing = False
            else:
                self.clearScale()
                self.drawing = False        
        else:
            return

    def rectChange(self, pos):
        if self.viewer._photo.isUnderMouse() and self.select_contours.isChecked():
            xi, yi, xf, yf = (self.click_x, self.click_y, pos.x(), pos.y())
            w,h = (abs(xf-xi), abs(yf-yi))
            try:
                if (w*h) <= self.Amin:
                    return
            except TypeError:
                if (w*h) <= 4:
                    return

            if self.viewer.button == 1: # Rubberband addition
                selected = RectangleOverlapTest(image=self.cv2_image, contours=self.large_contours,
                                                x=min(xi, xf), y=min(yi, yf), width=w, height=h)
                for c in selected:
                    self.contour_DF = self.contour_DF.append({
                            "uuid4": uuid.uuid4().hex, 
                            "contour": c, 
                            "C": self.C.value(), 
                            "kBlur": self.k_blur.value(),
                            "blocksize": self.blocksize.value(), 
                            "kLaplacian": self.k_laplacian.value(),
                            "kDilate": self.k_dilate.value(), 
                            "kGradient": self.k_gradient.value(), 
                            "kForeground": self.k_foreground.value()}, ignore_index=True)

            elif self.viewer.button == 2: # Rubberband subtraction
                to_remove = RectangleOverlapTest(image=self.cv2_image, contours=self.contour_DF["contour"].values,
                                                        x=min(xi, xf), y=min(yi, yf), width=w, height=h)
                self.contour_DF = self.contour_DF[~self.contour_DF["contour"].isin(to_remove)]
                self.contour_DF = self.contour_DF.reset_index(drop=True)
            else: return
            
            self.update_plot()

    def selectContoursChecked(self):
        if self.select_contours.isChecked():
            self.viewer.selectContoursChecked = True
            for slider in self.sliders:
                slider.setEnabled(False)
        else:
            self.viewer.selectContoursChecked = False
            for slider in self.sliders:
                slider.setEnabled(True)

    def resizeEvent(self, event):
        self.update_plot()

    def update_plot(self):
        # Clear image
        QPixmapCache.clear()
        # If actively tuning contours...
        self.cv2_image = cv2.imread(self.denoised_path.as_posix())
        if not self.select_contours.isChecked():
            self.contours = mcf(image=self.cv2_image,
                                k_blur=2 * self.k_blur.value() - 1,
                                C=self.C.value(),
                                blocksize=2 * self.blocksize.value() - 1,
                                k_laplacian=2 * self.k_laplacian.value() - 1,
                                k_dilate=2 * self.k_dilate.value() - 1,
                                k_gradient=2 * self.k_gradient.value() - 1,
                                k_foreground=2 * self.k_foreground.value() - 1, skip_flood=True)
            # Set min/max contour size
            try: Amin = int(self.Amin.text())
            except ValueError: Amin = int(1)
            try: Amax = int(self.Amax.text())
            except ValueError: Amax = int(self.cv2_image.shape[0]* self.cv2_image.shape[1])
            # Refine contours
            self.large_contours = contour_size_selection(self.contours, Amin=Amin, Amax=Amax)
            if self.use_approxPolys.isChecked():
                self.large_contours = [cv2.approxPolyDP(curve=c, epsilon=self.epsilon.value(), closed=True) for c in self.large_contours]
            if self.use_convexHulls.isChecked():
                self.large_contours = [cv2.convexHull(c) for c in self.large_contours]

        cv2.drawContours(self.cv2_image, contours=self.large_contours, contourIdx=-1, color=self.contour_color, thickness=self.contour_thickness.value())
        # Add scale bar
        if self.scaleBar is not None:
            if self.drawing:
                cv2.circle(self.cv2_image, (self.scaleBar[0], self.scaleBar[1]), radius=5, color=(0,255,0), thickness=-1)
                if len(self.scaleBar) == 4:
                    cv2.circle(self.cv2_image, (self.scaleBar[2], self.scaleBar[3]), radius=5, color=(0,255,0), thickness=-1)
                    cv2.line(self.cv2_image, (self.scaleBar[0], self.scaleBar[1]), (self.scaleBar[2], self.scaleBar[3]), (0,255,0),5)
            else:
                cv2.line(self.cv2_image, (self.scaleBar[0], self.scaleBar[1]), (self.scaleBar[2], self.scaleBar[3]), (0,255,0),5)

        if len(self.contour_DF) > 0:
            cv2.drawContours(self.cv2_image, contours=self.contour_DF["contour"].values, contourIdx=-1, color=self.highlight_color, thickness=self.contour_thickness.value())

        # Convert to QImage
        self.pixmap = cv2pixmap(self.cv2_image)
        self.viewer.setPhoto(self.pixmap)
        self.viewer.updateView()

    def update_contour_color(self):
        self.contour_color = QColorDialog.getColor().getRgb()
        self.update_plot()

    def update_highlight_color(self):
        self.highlight_color = QColorDialog.getColor().getRgb()
        self.update_plot()

    def update_text(self):
        self.blur_label.setText('k_blur: {}'.format(2*self.k_blur.value()-1))
        self.C_label.setText('C: {}'.format(self.C.value()))
        self.blocksize_label.setText('blocksize: {}'.format(2*self.blocksize.value()-1))
        self.laplacian_label.setText('k_laplacian: {}'.format(2*self.k_laplacian.value()-1))
        self.dilate_label.setText('k_dilate: {}'.format(2*self.k_dilate.value()-1))
        self.gradient_label.setText('k_gradient: {}'.format(2*self.k_gradient.value()-1))
        self.foreground_label.setText('k_foreground: {}'.format(2*self.k_foreground.value()-1))
        self.thickness_label.setText('Contour thickness: {}'.format(self.contour_thickness.value()))
        self.epsilon_label.setText('epsilon: {}'.format(self.epsilon.value()))

    def on_CLIgenerator_clicked(self):
        try: Amin = int(self.Amin.text())
        except ValueError: Amin = int(1)
        try: Amax = int(self.Amax.text())
        except ValueError: Amax = int(10e6)
        self.box.clear()
        self.box.insertPlainText("--k_blur {} --C {} --blocksize {} --k_laplacian {} --k_dilate {} --k_gradient {} --k_foreground {} --Amin {} --Amax {}".format(
                2 * self.k_blur.value() - 1,
                self.C.value(),
                2 * self.blocksize.value() - 1,
                2 * self.k_laplacian.value() - 1,
                2 * self.k_dilate.value() - 1,
                2 * self.k_gradient.value() - 1,
                2 * self.k_foreground.value() - 1,
                Amin, Amax))

    def copy_CLI(self):
        cb = QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText(self.box.toPlainText(), mode=cb.Clipboard)
        pyperclip.copy(self.box.toPlainText())

    def on_reset_clicked(self):
        self.k_blur.setValue(5)
        self.C.setValue(3)
        self.blocksize.setValue(8)
        self.k_laplacian.setValue(3)
        self.k_dilate.setValue(3)
        self.k_gradient.setValue(2)
        self.k_foreground.setValue(4)
        self.epsilon.setValue(3)
        self.Amin.clear()
        self.Amax.clear()

class GraphicsView(QtWidgets.QGraphicsView):
    photoClicked = QtCore.pyqtSignal(QtCore.QPoint)
    rectChanged = QtCore.pyqtSignal(QtCore.QPoint)

    def __init__(self, parent):
        super().__init__(parent)    
        self._zoom = 0
        self.factor = 1.0
        self._empty = True
        self._scene = QtWidgets.QGraphicsScene(self)
        self._photo = QtWidgets.QGraphicsPixmapItem()
        self._scene.addItem(self._photo)
        self.setScene(self._scene)
        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.setResizeAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.setBackgroundBrush(QtGui.QBrush(QtGui.QColor(30, 30, 30)))
        self.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.setCursor(QtCore.Qt.CrossCursor)
        self.button = None

        # Set up rubberband selection
        self.rubberband = QRubberBand(QRubberBand.Rectangle, self)
        self.setMouseTracking(True)
        self.origin = QtCore.QPoint()
        self.changeRubberBand = False

    def updateView(self):
        if self._zoom <= 1.0:
            scene = self.scene()
            r = scene.sceneRect()
            self.fitInView(r, QtCore.Qt.KeepAspectRatio)

    def resizeEvent(self, event):
        self.updateView()

    def setPhoto(self, pixmap=None):
        if pixmap and not pixmap.isNull():
            self._empty = False
            self._photo.setPixmap(pixmap)
        else:
            self._empty = True
            self._photo.setPixmap(QtGui.QPixmap())

    def keyPressEvent(self, event):
        zoom_key = False
        if event.key() == QtCore.Qt.Key_Equal and event.modifiers() == QtCore.Qt.ControlModifier:
            self.factor = 1.1
            self._zoom += 1
            zoom_key = True
        elif event.key() == QtCore.Qt.Key_Minus and event.modifiers() == QtCore.Qt.ControlModifier:
            self.factor = 0.9
            self._zoom -= 1
            zoom_key = True
        elif event.key() == QtCore.Qt.Key_Space and event.modifiers() == QtCore.Qt.ControlModifier:
            self.factor = 1.0
            self._zoom = 0
            self.updateView()
            zoom_key = True
        if zoom_key:
            if self._zoom > 0:
                self.scale(self.factor, self.factor)
            elif self._zoom == 0:
                self.updateView()

    def mousePressEvent(self, event):
        self.button = event.button()
        if self._photo.isUnderMouse():
            self.origin = event.pos()
            self.rubberband.setGeometry(QtCore.QRect(self.origin, QtCore.QSize()))
            self.changeRubberBand = True
            self.photoClicked.emit(self.mapToScene(event.pos()).toPoint())
        super(GraphicsView, self).mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if self.changeRubberBand:
            self.rubberband.show()
            self.rubberband.setGeometry(QtCore.QRect(self.origin, event.pos()).normalized())
            QGraphicsView.mouseMoveEvent(self, event)
        super(GraphicsView, self).mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if event.button() in [QtCore.Qt.LeftButton, QtCore.Qt.RightButton]:
            self.changeRubberBand = False
            if self.rubberband.isVisible():
                self.rectChanged.emit(self.mapToScene(event.pos()).toPoint())
                self.rubberband.hide()

        super(GraphicsView, self).mouseReleaseEvent(event)

def main():
    print("""
   __  ___   _          _               
  /  |/  /  (_)  ___   (_)              
 / /|_/ /  / /  / _ \ / /               
/_/__/_/  /_/  /_//_//_/                 
 / ___/ ___   ___  / /_ ___  __ __  ____
/ /__  / _ \ / _ \/ __// _ \/ // / / __/
\___/__\___//_//_/\__/ \___/\_,_/ /_/   
  / __/  (_)  ___  ___/ / ___   ____    
 / _/   / /  / _ \/ _  / / -_) / __/    
/_/    /_/  /_//_/\_,_/  \__/ /_/       
                                        
    """)

    app = QApplication(sys.argv)
    w = MainWindow()
    w.show()
    app.exec_()

if __name__ == '__main__':
    main()


#!/usr/bin/env python
# coding: utf-8

# core
import sys, os, json, codecs, uuid
import numpy as np
from datetime import datetime
import pandas as pd
from scipy import spatial
from tqdm import tqdm
from pathlib import Path, PosixPath
from typing import Union
# plotting
import matplotlib
if sys.platform == 'darwin':
    matplotlib.use("TkAgg")
import matplotlib.pyplot as plt
from matplotlib.offsetbox import OffsetImage, AnnotationBbox
# image related
import cv2
from PIL import Image
from pytesseract import image_to_string
from PyQt5.QtGui import QPixmap, QImage

# Custom utils
try:
    from helpers import *
except ModuleNotFoundError:
    from MCF.helpers import *

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def flood_fill(image, progress_bar=True):
    """Flood fill from the edges."""
    height, width = image.shape[:2]
    for row in range(height):
        if image[row, 0] == 255:
            cv2.floodFill(image, None, (0, row), 0)
        if image[row, width-1] == 255:
            cv2.floodFill(image, None, (width-1, row), 0)
    if progress_bar:
        for col in tqdm(range(width), desc='Flooding background', leave=False):
            if image[0, col] == 255:
                cv2.floodFill(image, None, (col, 0), 0)
            if image[height-1, col] == 255:
                cv2.floodFill(image, None, (col, height-1), 0)
    else:
        for col in range(width):
            if image[0, col] == 255:
                cv2.floodFill(image, None, (col, 0), 0)
            if image[height-1, col] == 255:
                cv2.floodFill(image, None, (col, height-1), 0)

    return image

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def mcf(image, k_blur=9, C=3, blocksize=15, k_laplacian=5, k_dilate=5, k_gradient=3, k_foreground=7,
        extract_border=False, offsetX=0, offsetY=0, skip_flood=False, debug=False, progress_bar=True):
    """Extracts contours from an image. Can be used to extract a contour surrounding the
    entire foreground border.

    Parameters
    ----------
    image : <numpy.ndarray> Query image
    k_blur : <int> 9; blur kernel size; must be odd
    C : <int> 3; constant subtracted from mean during adaptive Gaussian smoothing
    blocksize : <int> 15; neighborhood size for calculating adaptive Gaussian threshold; must be odd
    k_laplacian : <int> 5; laplacian kernel size; must be odd
    k_dilate : <int> 5; dilation kernel size; must be odd
    k_gradient : <int> 3; gradient kernel size; must be odd
    k_foreground : <int> 7; foregound clean up kernel size; must be odd
    extract_border : <bool> False; extract background o
    debug : <bool> writes debugging information and plots each step

    Returns
    -------
    contours : <list> A list of contours"""
    """Gray"""
    if debug: print("[PID {}] Gray...".format(os.getpid()))
    gray = cv2.cvtColor(src = image, code=cv2.COLOR_RGB2GRAY)
    """Adaptive histogram normalization"""
    gridsize = int(min(0.01*max(image.shape[:2]), 8))
    clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(gridsize, gridsize))
    hequalized = clahe.apply(gray)
    """Blur"""
    if debug: print("[PID {}] Blur...".format(os.getpid()))
    blur = cv2.GaussianBlur(src=hequalized, ksize=(k_blur, k_blur), sigmaX=2, )
    """Adaptive Gaussian threshold"""
    if debug: print("Adapt Gauss Thresh...")
    thresh = cv2.adaptiveThreshold(src=blur, maxValue=255, adaptiveMethod=cv2.ADAPTIVE_THRESH_GAUSSIAN_C, thresholdType=cv2.THRESH_BINARY, blockSize=blocksize, C=C)
    """Laplacian"""
    if debug: print("Laplacian...")
    laplacian = cv2.Laplacian(src=thresh, ddepth=cv2.CV_16S, ksize=k_laplacian, )
    """Dilate"""
    if debug: print("Dilate...")
    kernel = cv2.getStructuringElement(shape=cv2.MORPH_RECT, ksize=(k_dilate, k_dilate))
    dilate = cv2.dilate(laplacian, kernel=kernel, iterations=1)
    """Morphological gradient"""
    if debug: print("Gradient...")
    kernel = cv2.getStructuringElement(shape=cv2.MORPH_RECT, ksize=(k_gradient, k_gradient))
    gradient = cv2.morphologyEx(dilate, cv2.MORPH_GRADIENT, kernel=kernel, iterations=1)
    """Binarize"""
    if debug: print("Binarize...")
    tozero = cv2.threshold(gradient, 127, 255, cv2.THRESH_TOZERO)
    tozero = np.uint8(np.uint8(tozero[1]))
    binary = cv2.inRange(tozero, 0, 100)
    """Foreground clean up"""
    if debug: print("Foreground...")
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (k_foreground, k_foreground))
    foreground = cv2.morphologyEx(binary, cv2.MORPH_OPEN, kernel, iterations=1)
    foreground = cv2.morphologyEx(foreground, cv2.MORPH_CLOSE, kernel, iterations=1)
    """Flood from outside"""
    if debug: print("Flood fill...")
    if skip_flood:
        flood = foreground.copy()
    else:
        flood = flood_fill(foreground, progress_bar=progress_bar)
    if extract_border:
        """Get border"""
        if debug: print("Getting border...")
        kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (17, 17))
        background = cv2.dilate(flood, kernel, iterations=2)
        background_contours, _ = cv2.findContours(image=background.copy(), mode=cv2.RETR_EXTERNAL, method=cv2.CHAIN_APPROX_SIMPLE)
        border_contour = [max(background_contours, key=cv2.contourArea)]
        return border_contour

    """Find contours"""
    if debug: print("Drawing contours...")
    contours, _ = cv2.findContours(image=flood.copy(), mode=cv2.RETR_EXTERNAL, method=cv2.CHAIN_APPROX_SIMPLE)
    return [c + [offsetX, offsetY] for c in contours]

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def parallel_mcf(window, **kwargs):
    return mcf(image=window[2], offsetX=window[0], offsetY=window[1], progress_bar=False, **kwargs)

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def contour_size_selection(contours, Amin=50, Amax=10e6):
    """Selections contours based on perimeter and area

    Parameters
    ----------
    contours : <list> A list of contours
    Amin : <int> 50; Minimum area in pixels
    Amax : <int> 10e6; Maximum area in pixels

    Returns
    -------
    contours : <list> A list of contours
    """

    return [c for c in contours if Amax >= cv2.contourArea(c) >= Amin]

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def smooth_contours(contours, epsilon=1):
    """Returns slightly smoothed and convex hulls versions of the input contours.

    Parameters
    ----------
    contours : <list> A list of contours

    Returns
    -------
    smoothed : <list> A list of smoothed contours
    hulls : <list> A list of contour convex hulls
    """
    smoothed = []
    hulls = []
    for c in contours:
        smoothed.append(cv2.approxPolyDP(curve=c, epsilon=epsilon, closed=True))
        hulls.append(cv2.convexHull(c))

    return smoothed, hulls

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def sliding_window(image, stepSize, windowSize):
    """Creates a generator of overlapping windows that tile across an image.

    Parameters
    ----------
    image : <numpy.ndarray> Query image
    stepsize : <int> Slide step size in pixels (currently the same in x and y directions)
    windowSize: <int>

    Returns
    -------
    smoothed : <list> A list of smoothed contours
    hulls : <list> A list of contour convex hulls
    """
    # slide a window across the image
    for y in range(0, image.shape[0], stepSize):
        for x in range(0, image.shape[1], stepSize):
            # yield the current window
            yield (x, y, image[y:y + windowSize[1], x:x + windowSize[0]])

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def sliding_contour_finder(image, stepsize, winW, winH, border_contour, skip_flood=False, debug=False, **kwargs):
    """Uses a sliding-window approach to find contours across a large image. Uses KDTree algorithm to
    remove duplicated contours from overlapping windows.

    Parameters
    ----------
    image : <numpy.ndarray> Query image
    stepsize : <int> Slide step size in pixels (currently the same in x and y directions)
    winW : <int> Window width in pixels
    winH : <int> Window height in pixels
    **kwargs : Kwargs passed to `mcf`

    Returns
    -------
    contours : <list> A list of contours
    smooth_contours : <list> A list of smoothed contours"""

    """Create windows for mini contour finder"""
    if debug: print("Creating windows...")

    # Create image of border
    clone = image.copy()
    blank = np.zeros(clone.shape[0:2], dtype=np.uint8)
    border_mask = cv2.drawContours(blank.copy(), border_contour, 0, (255), -1)
    # mask input image (leaves only the area inside the border contour)
    cutout = cv2.bitwise_and(clone, clone, mask=border_mask)

    n_windows = len(list(sliding_window(image=cutout.copy(), stepSize=stepsize, windowSize=(winW, winH))))
    windows = sliding_window(image=cutout.copy(), stepSize=stepsize, windowSize=(winW, winH))

    contours = []
    for i, (x,y,window) in tqdm(enumerate(windows), total=n_windows, desc='Windows'):
        if debug: print(("Window {}, x0: {}, y0: {}, shape: {}".format(i,x,y,np.shape(window))))
        if window.shape[0] != winH or window.shape[1] != winW: continue
        if window.sum() == 0: continue
        """Running mini contour finder in window"""
        if debug: print("Running mini contour finder...")
        contours += mcf(window, skip_flood=skip_flood, progress_bar=False, offsetX=x, offsetY=y, **kwargs)

    return contours

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def remove_redundant_contours(contours, neighborhood=10):
    moments = []
    nonredundant = []
    for c in contours:
        cX,cY = contour_xy(c)

        if len(moments)==0:
            nonredundant.append(c)
            moments.append([cX, cY])
        else: # if previous moments exist, find the distance and index of the nearest neighbor
            distance, _ = spatial.KDTree(moments).query([cX, cY])
            if distance > neighborhood: # add point if moment falls outside of neighborhood
                nonredundant.append(c)
                moments.append([cX, cY])

    return nonredundant

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def getContourRBG(contour, image):
    clone = image.copy()
    mask = np.zeros_like(clone)  # Create mask where white is what we want, black otherwise
    cv2.drawContours(mask, contour, 0, 255, -1)  # Draw filled contour in mask
    out = np.zeros_like(image)  # Extract out the object and place into output image
    out[mask == 255] = image[mask == 255]
    pixelpoints = np.transpose(np.nonzero(mask))

    return image[flatten(pixelpoints[:, :2])[::2], flatten(pixelpoints[:, :2])[1::2]].mean(axis=0) / 255

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def contour_xy(contour):
    M = cv2.moments(contour)
    if M["m00"] != 0:
        return (int((M["m10"] / M["m00"]))-10, int((M["m01"] / M["m00"])))
    else: 
        return (0,0)

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def export_contour_data(contourDF:pd.DataFrame, prefix:str, image:np.ndarray, conversion_factor=None, units=None, output_dir="./", get_color=False):
    """Exports contours, contour metadata, and summary statistics.

    Parameters
    ----------
    image : <np.ndarray> Image
    contourDF : <np.ndarray> Contours to export
    conversion_factor : <float> Conversion factort to convert pixel length. Default=None
    units : <str> Associated conversion factor units. Default=None
    prefix : <str> Prefix for output files (e.g. PREFIX.contour_data.pkl, PREFIX.contour_data.csv)
    output_dir : <str> Path to output directory. Default='./' """
    
    contourDF["area_pixels"] = np.array(list(map(cv2.contourArea, contourDF["contour"].values)), dtype=object)
    contourDF["moment_XY"] = [contour_xy(c) for c in contourDF["contour"].values]
    cRXY = np.array(list(map(cv2.minEnclosingCircle, contourDF["contour"].values)), dtype=object)
    contourDF["min_circle_xy"] = cRXY[:,0]
    contourDF["min_circle_r"] = cRXY[:,1]
    bbox = np.array([cv2.boundingRect(c) for c in contourDF["contour"].values])
    contourDF["bbox_x"], contourDF["bbox_y"], contourDF["bbox_h"], contourDF["bbox_w"] = bbox[:,0], bbox[:,1], bbox[:,2], bbox[:,3]
    contourDF["bbox_area"] = contourDF["bbox_w"]*contourDF["bbox_h"]
    contourDF["aspect_ratio"] = contourDF["bbox_w"]/contourDF["bbox_h"]
    contourDF["convex_hull"] = [cv2.convexHull(c) for c in contourDF["contour"].values]
    contourDF["convexity"] = [cv2.isContourConvex(c) for c in contourDF["contour"].values]
    contourDF["solidity"] = [float(cv2.contourArea(c))/cv2.contourArea(cv2.convexHull(c)) for c in contourDF["contour"].values]
    contourDF["equivalent_diameter"] = [np.sqrt(4*cv2.contourArea(c)/np.pi) for c in contourDF["contour"].values]

    if conversion_factor and units:
        contourDF["area_{}^2".format(units)] = contourDF["area_pixels"]*(conversion_factor**2)

    if get_color:
        if type(image)==str:
            image = cv2.imread(image)
        contourDF["RBG"] = np.array(list(map(getContourRBG, contourDF["contour"].values)), dtype=object)

    # Export
    contourDF.to_csv(Path(output_dir) / "{}.contour_data.csv".format(prefix), index=False)
    contourDF.to_json(Path(output_dir) / "{}.contour_data.json".format(prefix))

    print("[{}] Contour data exported to".format(datetime.now().strftime('%d %b %Y %H:%M:%S')))
    print("\t{}".format(Path(output_dir) / "{}.contour_data.csv".format(prefix)))
    print("\t{}".format(Path(output_dir) / "{}.contour_data.json".format(prefix)))
   
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def render_contour_plots(image, border_contour, contours, prefix, dpi=300, output_dir="./", color = (0, 255, 255), contour_thickness=3, imgFormat="pdf"):
    """Creates two contour plots: 1) border and interior contours overlaid on image 2) border and interior
    contours overlaid on image with contour indices for reference.

    Parameters
    ----------
    img : <numpy.ndarray> Query image
    border_contour : <np.ndarray> Border contour to plot
    contours : <np.ndarray> Interior contours to plot
    moments : <list> List of contour moments
    prefix : <str> Prefix for output files (e.g. prefix.noindex.pdf, prefix.pdf)
    dpi : <int> Output image resolution. Default=300
    output_dir : <str> Path to output directory. Default='./'
    """
    matplotlib.use("PDF")
    # Get figure size
    if type(image)==str:
        try:
            image = cv2.imread(image)
            image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        except cv2.error:
            image = vector2cv2(image.as_posix())

    height, width, _ = image.shape
    figsize = width / float(dpi), height / float(dpi)

    # No index image
    canvas = image.copy()
    _, ax = plt.subplots(ncols=1, figsize=figsize)
    if border_contour:
        cv2.drawContours(canvas, contours=border_contour, contourIdx=-1, color=(255,0,0), thickness=contour_thickness)
    cv2.drawContours(canvas, contours=contours, contourIdx=-1, color=color, thickness=contour_thickness)
    ax.imshow(canvas)
    ax.axis('off')
    plt.savefig(Path(output_dir) / "{}.noindex.{}".format(prefix,imgFormat), dpi=dpi, transparent=True)
    plt.close()

    # Indexed image
    canvas = image.copy()
    _, ax = plt.subplots(ncols=1, figsize=figsize)
    if border_contour:
        cv2.drawContours(canvas, contours=border_contour, contourIdx=-1, color=(255,0,0), thickness=contour_thickness)
    # Plot indexed contours
    cv2.drawContours(canvas, contours=contours, contourIdx=-1, color=color, thickness=contour_thickness)
    for i, c in enumerate(contours):
        cX,cY = contour_xy(c)
        ax.text(x=cX, y=cY, s=u"{}".format(i), color="black", size=5)

    ax.imshow(canvas)
    ax.axis('off')
    plt.savefig(Path(output_dir) / "{}.{}".format(prefix, imgFormat), dpi=dpi, transparent=True)
    plt.close()
    print("[{}] Contour plots saved to".format(datetime.now().strftime('%d %b %Y %H:%M:%S')))
    print("\t{}".format(Path(output_dir) / "{}.noindex.{}".format(prefix, imgFormat)))
    print("\t{}".format(Path(output_dir) / "{}.index.{}".format(prefix, imgFormat)))
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def process_image(image_path, neighborhood=10, prefix=None, Amin=50, Amax=10e6, detectScaleBar=False,
                  output_dir="./", print_plots=True, dpi=300, debug=False, **kwargs):
    """Parameters
    ----------
    image <str> or <numpy.ndarray> : Query image. If string, is assumed to be an image filepath; if numpy.ndarray,
        assumed to be in cv2 or numpy format.
    Amin <int> : Minimum contour area in pixels
    Amax <int> : Maximum contour area in pixels
    sliding_window <bool> : Use sliding window contour detection. Default=True
    neighborhood <int> : Neighborhood size in pixels determining a unique contour
    prefix <str> : New prefix for output files. By default the new files will reflect the input file's basename
    output_dir <str> : Path to output directory. Default='./'
    debug <bool> : writes debugging information and plots each step
    **kwargs : kwargs for `mcf`
    """

    if debug:
        print("[{}] Working directory: {}".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), os.getcwd()))
        print("[{}] Output directory: {}".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), output_dir))
    input_path = Path(image_path)
    image = cv2.imread(input_path.as_posix())
    if image == None:
        image = vector2cv2(image_path.as_posix())

    if not prefix:
        prefix = input_path.stem
    if debug:
        print("[{}] Input file: {}".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), input_path.absolute()))

    """Denoise"""
    wdir = image_path.parent
    results = wdir.glob("{}.denoise.*".format(image_path.stem))
    try:
        denoised_path = next(results)
        denoise = cv2.imread(denoised_path.as_posix())
        print("[{}] Found existing denoised file ({})".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), denoised_path.as_posix()))
    except StopIteration:
        print("[{}] Denoising image...".format(datetime.now().strftime('%d %b %Y %H:%M:%S')))
        denoise = cv2.fastNlMeansDenoisingColored(image, None, 10, 10, 7, 21)
        denoised_path = Path("{}/{}.denoise{}".format(wdir.as_posix(), image_path.stem, ".png"))
        cv2.imwrite(filename=denoised_path.as_posix(), img=denoise)
        print("[{}] Created temporary denoised file ({})".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), denoised_path.as_posix()))

    if detectScaleBar:
        scaleBarInfo = get_scalebar_info(denoise)
        if len(scaleBarInfo) == 2:
            scaleBar, scaleBarPixelLength = scaleBarInfo
            scaleBarUnits = None
            unitsPerPixel= None
            print("[{}] Detected scale bar but could not read units. Try detection or manual drawing in MCF-GUI.".format())
            return
        elif len(scaleBarInfo) == 4:
            scaleBar, scaleBarPixelLength, scaleBarUnitLength, scaleBarUnits = scaleBarInfo
            unitsPerPixel = scaleBarUnitLength/scaleBarPixelLength
        else:
            scaleBarUnits = None
            unitsPerPixel = None
            print("[{}] Could not find scale bar. Try detection or manual drawing in MCF-GUI.".format())
            return
        cv2.line(image, (scaleBar[0], scaleBar[1]), (scaleBar[2], scaleBar[3]), (0,255,0),5)
    else:
        scaleBarUnits = None
        unitsPerPixel = None

       
    print("[{}] Getting image border...".format(datetime.now().strftime('%d %b %Y %H:%M:%S')))
    border_contour = mcf(image=denoise, extract_border=True, **kwargs)

    print("[{}] Finding contours...".format(datetime.now().strftime('%d %b %Y %H:%M:%S')))
    contours = mcf(denoise, skip_flood=False, progress_bar=True, **kwargs)
    contours = contour_size_selection(contours, Amin=Amin, Amax=Amax)
    contours = remove_redundant_contours(contours, neighborhood=neighborhood)
    contour_DF = pd.DataFrame(columns=["uuid4", "contour", "C", "kBlur", "blocksize", "kLaplacian", "kDilate", "kGradient", "kForeground"], dtype=object)
    contour_DF["contour"] = contours
    contour_DF["uuid4"] = [uuid.uuid4().hex for i in range(len(contour_DF))]
    contour_DF["C"] = kwargs['C']
    contour_DF["kBlur"] = kwargs['k_blur']
    contour_DF["blocksize"] = kwargs['blocksize']
    contour_DF["kLapacian"] = kwargs['k_laplacian']
    contour_DF["kDilate"] = kwargs['k_dilate']
    contour_DF["kGradient"] = kwargs['k_gradient']
    contour_DF["kForeground"] = kwargs['k_foreground']
    print("[{}] Found {} contours".format(datetime.now().strftime('%d %b %Y %H:%M:%S'), len(contours)))
    print("[{}] Exporting contour data...".format(datetime.now().strftime('%d %b %Y %H:%M:%S')))
    # Export contours
    output_dir = Path(output_dir)
    if not output_dir.is_dir():
        output_dir.mkdir(parents=True, exist_ok=True)
    if unitsPerPixel and scaleBarUnits:
        export_contour_data(image=image, contourDF=contour_DF, conversion_factor=unitsPerPixel,
                        units=scaleBarUnits, prefix=prefix, output_dir=output_dir, get_color=False)
    else:
        export_contour_data(image=image, contourDF=contour_DF, prefix=prefix, output_dir=output_dir, get_color=False)
    if print_plots:
        print("[{}] Plotting...".format(datetime.now().strftime('%d %b %Y %H:%M:%S')))
        render_contour_plots(image=image, border_contour=border_contour, contours=contour_DF['contour'].values, 
        prefix=prefix, dpi=dpi, output_dir=output_dir)


#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def detect_scalebar(image, rho=1, theta=float(np.pi/180), min_votes=15, min_line_length=100,
                 max_line_gap=0, threshold1=50, threshold2=150, largest=True):
    """Detects line-like (i.e. a straight bar instead of a ruler) scalebars.
    Written by StackOverflow user SHEN and amended by Ian Gilman
    https://stackoverflow.com/a/45560545/6317380

    Parameters
    ----------
    rho : <float> Distance resolution in pixels of the Hough grid. Default = 1
    theta : <float> Angular resolution in radians of the Hough grid. Default = np.pi / 180
    min_votes : <int> Minimum number of votes (intersections in Hough grid cell). Default = 15
    min_line_length : <int> Minimum number of pixels making up a line. Default = 100
    max_line_gap : <int> Maximum gap in pixels between connectable line segments. Default = 0
    threshold1 : <int> First threshold for the hysteresis procedure. Default = 50
    threshold2 : <int> Second threshold for the hysteresis procedure. Default = 150
    color : <tuple> Color as BGR tuple for plotting. Default = (255,0,0)
    largest : <bool> Only return largest scalebar

    Returns
    -------
    line_image : <numpy.ndarray> Resulting image of detected scalebar lines
    """

    line_image = image.copy()*0  # creating a blank to draw lines on
    edges = cv2.Canny(image.copy(), threshold1=threshold1, threshold2=threshold2)

    # Run Hough on edge detected image
    # Output "lines" is an array containing endpoints of detected line segments
    lines = cv2.HoughLinesP(image=edges.copy(), rho=rho, theta=theta, threshold=min_votes,
                            lines=np.array([]), minLineLength=min_line_length, maxLineGap=max_line_gap)

    scalebars = np.array([0,0,0,0])
    scalebar_length = 0.0
    # Get largest scalebar
    if largest:
        for l in lines:
            x1 = l[0][0]
            x2 = l[0][2]
            y1 = l[0][1]
            y2 = l[0][3]

            length = np.sqrt((x2-x1)**2+(y2-y1)**2)

            if length>scalebar_length:
                scalebar_length = length
                scalebars = l[0]
        # Add scalebar to image
        cv2.line(line_image,(x1,y1),(x2,y2),(255,0,0),5)
    else:
        scalebars = lines.copy()
        length = None
        for line in np.array(scalebars):
            for x1,y1,x2,y2 in line:
                cv2.line(line_image,(x1,y1),(x2,y2),(255,0,0),5)

    return line_image, scalebars, length

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def read_units(scalebar_img):
    """Reads the units and metric of scale bar from an image and uses the number
    of pixels in the scale bar to convert pixel area to units detected in the image.

    Parameters
    ----------
    scalebar : <list>, Scale bar endpoints [x1, y1, x2, y2]
    scarebar_img : <numpy.ndarray>, Image of scale bar

    Returns
    -------
    pixel_area : <float>, The area of a single pixel in the new units
    converstion_units : <str>, The detected units"""

    scalebar_length, scalebar_units = image_to_string(image=scalebar_img.copy(), lang="eng").split()

    # Check for units in micrometers
    eng_units = image_to_string(image=scalebar_img.copy(), lang='eng').split()[1]
    grc_units = image_to_string(image=scalebar_img.copy(), lang='grc').split()[1]
    if (grc_units[0]==u"\u03BC") and (eng_units in ["um", "pm"]):
        scalebar_units = grc_units[0]+eng_units[1:]

    scalebar_length = float(scalebar_length)

    return scalebar_length, scalebar_units

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def vector2cv2(file_path):
    """Converts a vector image to a BGR cv2 format."""
    return np.array(Image.open(file_path))[:,:,::-1]

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_scalebar_info(image, plot=False, **kwargs):
    """Detects and reads a scalebar.

    Parameters
    ----------
    image : <np.ndarray> Image
    plot : <bool> Plot resulting scalebar. Default = False
    **kwargs : Kwargs to pass to `detect_scalebar`

    Returns
    -------
    conversion_factor : <float> Conversion factor from pixel area to new unit area (e.g. 0.1 um^2/pixel)
    units : <str> New units"""

    clone = image.copy()
    # Detect largest scalebar
    try:
        line_image, scalebar, length_in_pixels = detect_scalebar(clone, **kwargs)
    except TypeError:
        print("Could not detect scalebar.")
        return
    # Add scalebar to image
    line_edges = cv2.addWeighted(src1=clone, alpha=1.0, src2=line_image.copy(), beta=10, gamma=0)
    height, width = line_edges.shape[:2]

    # Easiest to read scalebar text when isolated. Start with small window around scalebar.
    pad = 50
    while pad <= max(height, width):
        try:
            xmin = scalebar[0]
            xmax = scalebar[2]
            ymin = scalebar[1]
            ymax = scalebar[3]
            crop_scalebar = line_edges.copy()[max([0, ymin-pad]):min([height, ymax+pad]), max([0, xmin-pad]):min([width, xmax+pad])]
            length_in_units, scalebar_units = read_units(scalebar_img=crop_scalebar)
            if plot:
                fig, ax = plt.subplots(figsize=(6,6))
                ax.imshow(crop_scalebar)
                plt.tight_layout()

            return scalebar, length_in_pixels, length_in_units, scalebar_units

        # If no units were found, expand search window.
        except ValueError:
            pad = pad*2
    print("Detected scalebar but could not read units.")
    return scalebar, length_in_pixels
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def RectangleOverlapTest(image, contours, x, y, width, height, REMOVE=False):
    """Finds contours that overlap with a rectangle.

    Parameters
    ----------
    image <numpy.ndarray> : Image which contours are from
    contours <list>: A list of contours
    x <int> : Left rectangle side
    y <int> : Top rectangle side
    width <int> : Rectangle width
    height <int> : Rectangle height

    Returns
    -------
    selected <list> : A list of contours
    """
    blank = np.zeros(image.shape[0:2])
    rectangle = cv2.rectangle(blank.copy(), (x, y), (x + width, y + height), 1, cv2.FILLED)
    rect_x, rect_y = (x + (width / 2), y + (height / 2))
    if REMOVE:
        rm_idx = []
        for i, c in tqdm(enumerate(contours), desc='RectangleOverlapTest', leave=False):
            (min_c_x, min_c_y), min_c_r = cv2.minEnclosingCircle(c)
            if abs(min_c_x - rect_x) > (width / 2) + (min_c_r): continue
            if abs(min_c_y - rect_y) > (height / 2) + (min_c_r): continue

            current = cv2.drawContours(blank.copy(), contours, i, 1, cv2.FILLED)
            overlap = np.logical_and(current, rectangle)
            if overlap.any():
                rm_idx.append(i)
        
        return [i for j, i in enumerate(contours) if j not in rm_idx]

    else:
        selected = []
        for i, c in tqdm(enumerate(contours), desc='RectangleOverlapTest', leave=False):
            (min_c_x, min_c_y), min_c_r = cv2.minEnclosingCircle(c)
            if abs(min_c_x - rect_x) > (width / 2) + (min_c_r): continue
            if abs(min_c_y - rect_y) > (height / 2) + (min_c_r): continue

            current = cv2.drawContours(blank.copy(), contours, i, 1, cv2.FILLED)
            overlap = np.logical_and(current, rectangle)
            if overlap.any():
                selected.append(c)
        return selected

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def ContourOverlapTest(image, contours, background_contours, return_overlapping=True):
    """finds contours that overlap with a set of background contours.

    Parameters
    ----------
    image <numpy.ndarray> : Image which contours are from
    contours <list> : List of contours of interest
    background_contours  <list> : List of background contours to look for overlap with
    return_overlapping <bool> : Return overlapping contours. If False, will return contours that
        do not overlap. Default=True

    Returns
    -------
    selected <list> : A list of contours that are overlapping
    """
    selected = []
    blank = np.zeros(image.shape[0:2])
    background_image = cv2.drawContours(blank.copy(), background_contours, -1, 1, cv2.FILLED)
    for i in tqdm(range(len(contours)), desc='ContourOverlapTest', leave=False):
        current = cv2.drawContours(blank.copy(), contours, i, 1, cv2.FILLED)
        if return_overlapping:
            if np.logical_and(current, background_image).any():
                selected.append(contours[i])
        else:
            if not np.logical_and(current, background_image).any():
                selected.append(contours[i])

    return selected

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def cv2pixmap(image):
    """Creates a QPixmap object from a cv2 image"""
    if type(image) == str and Path(image).exists():
        return QPixmap(image)
    elif type(image) == np.ndarray:
        cv2_image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        h, w, _ = cv2_image.shape
        totalBytes = cv2_image.nbytes
        bytesPerLine = int(totalBytes / h)
        qimage_item = QImage(cv2_image.data, w, h, bytesPerLine, QImage.Format_RGB888)
        return QPixmap(qimage_item)
    else:
        print("ERROR: Could not parse image.")
        sys.exit()

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
class NumpyEncoder(json.JSONEncoder):
    """Written by StackOverflow user karlB
    https://stackoverflow.com/a/47626762/6317380"""
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def contour2json(contours: list, file_path: Union[str, PosixPath]):
    """Saves contours to json using NumpyEncoder"""
    if type(file_path) == 'pathlib.PosixPath':
        file_path = file_path.as_posix()
    
    json.dump(obj=contours, fp=codecs.open(file_path, 'w+', encoding="utf-8"), cls=NumpyEncoder)

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def read_json_contours(file_path: Union[str, PosixPath]) -> list:
    """Reads contours from a json"""
    if type(file_path) == 'pathlib.PosixPath':
        file_path = file_path.as_posix()
    
    json_load = json.load(codecs.open(file_path, 'r', encoding="utf-8"))

    return [np.array(c, dtype='int32') for c in json_load]

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def contourScatter(x, y, contours, color, linewidth=1, alpha=1.0, ax=None, zoom=1.0):
    """A modified scatter plot that allows plotting a set of points that
    define a line as a marker. Written by Stackoverflow user Joe Kington 
    and modified for contour use by HansHirse.
    https://stackoverflow.com/a/22570069/11089932

    Parameters
    ----------
    x <iterable> : x-positions
    y <iterable> : y-positions
    contours <iterable of numpy.ndarray> : Contours to plot as markers
    color <str, tuple, or iterable> : Matplotlib color string or color in RGB or RGBA 
        format, or an iterable of those types
    linewidth <int> : Contour line width. If -1, will fill shape with <color>
    alpha <float> : Marker transparency
    ax <matplotlib.axes> : Axes to use
    zoom <float> : Marker scale
    
    Returns
    -------
    artists <list> : List of plot attributes"""

    if ax is None:
        ax = plt.gca()
        
    x, y = np.atleast_1d(x, y)
    
    if type(color)==str or np.shape(color) in [(3,), (4,)]:
        color = np.reshape(np.tile(rgba2bgra(color=color), len(x)), (len(x),4))
    else:
        if len(color)==len(x):
            color = [rgba2bgra(c) for c in color]
        else:
            print("Length of color does not match number of contours.")
        color = [rgba2bgra(c) for c in color]

    artists = []
    for x0, y0, cont, col in zip(x, y, contours, color):
        cx, cy, cw, ch = cv2.boundingRect(cont)
        img = np.zeros((ch + 11, cw + 11, 4), np.uint8)
        img = cv2.drawContours(img, [cont], -1, col, linewidth, offset=(-cx+5, -cy+5))
        img = OffsetImage(img, zoom=zoom, alpha=alpha)
        ab = AnnotationBbox(img, (x0, y0), xycoords='data', frameon=False)
        artists.append(ax.add_artist(ab))
    ax.update_datalim(np.column_stack([x, y]))
    ax.autoscale()
    
    return artists

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def rgba2bgra(color):
    if type(color)==str: # String to BGRA255
        try:
            color = np.array(matplotlib.colors.to_rgba(color))[[2,1,0,3]]*255
        except ValueError:
            print("Invalid matplotlib color")
            return 
    elif np.shape(color) == (3,): #RGB to RGBA
        color = np.array(color)[::-1]*255
        color = np.append(color,255)
    elif np.shape(color) == (4,):
        color = np.array(color)[[2,1,0,3]]*255
    else:
        print("Invalid matplotlib color")
        return
    return color


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="MiniContourFinder",
    version="1.0.17",
    author="Ian S Gilman",
    author_email="ian.gilman@yale.edu",
    description="Lightweight image segmentation software for biological images",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/isgilman/MiniContourFinder",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Image Processing",
        "Intended Audience :: Science/Research",
    ],
    install_requires=[
        "matplotlib >= 3.1.1",
        "numpy > 1.20.3",
        "opencv-python >= 4.5.2",
        "pandas >= 0.25.2",
        "pillow >= 6.2.1",
        "Pillow",
        "pyperclip >= 1.7.0",
        "PyQt5 >= 5.12.3",
        "pytesseract >= 0.3.0",
        "scipy >= 1.6.3",
        "tqdm >= 4.36.1"
    ],
    entry_points={
        'console_scripts':[
            'mcf=MCF.mcf:main',
            'mcf_gui=MCF.mcf_gui:main',
            'mcf_parallel=MCF.mcf_parallel:main'
            ]}
)